* FIFO device driver for QIMSI
* This is a test version. Not to be used other than for testing purposes!
* v1.1 20231014 JB  Use STX1 and SRX1 as device names for SerNet

txdata0 equ     $fed2
txdata1 equ     $fed3
rxdata  equ     $fed4
status  equ     $fed5
txempty equ     0
rxempty equ     1
rxoverr equ     3

        include mt_inc
        include err_inc
        include vect_inc

; physical definition block

        offset  0

        ds.l    1       ; link to next driver
        ds.l    1       ; physical I/O routine (jsr io.serio)
        ds.l    1       ; open routine
        ds.l    1       ; close routine
        ds.l    1       ; JSR instruction to io.serio (short)
        ds.l    1       ; address of pending i/o routine
        ds.l    1       ; address of fetch byte routine
        ds.l    1       ; address of send byte routine
        ds.w    1       ; RTS instruction
phys_sz equ     *       ; length of physical definition block

        section dev_fifo

devinit:
        moveq   #phys_sz,d1     ; allocate physical def block
        moveq   #0,d2
        moveq   #mt.alchp,d0
        trap    #1
        tst.l   d0
        bne.s   initerr
        lea     dev_def,a3      ; set up pointers
        lea     8(a0),a2
        bsr.s   getaddr         ; open routine
        bsr.s   getaddr         ; close routine
        move.l  a2,4(a0)        ; physical i/o routine
        move.w  #$4eb8,(a2)+    ; jsr absolute short
        move.w  io.serio,(a2)+  ; to io_serio
        bsr.s   getaddr         ; pending i/o routine
        bsr.s   getaddr         ; fetch character
        bsr.s   getaddr         ; send character
        move.w  #$4e75,(a2)     ; rts
        moveq   #mt.liod,d0     ; link in driver
        trap    #1
initerr rts

* calculate and store the absolute addresses required for io_serio

getaddr move.w  (a3)+,a1                ; get relative pointer
        lea     -2(a3,a1.w),a1          ; make address absolute
        move.l  a1,(a2)+                ; and store it
        rts
                    
dev_def dc.w    open-*          ; open routine
        dc.w    close-*         ; close routine
        dc.w    pend-*          ; pending input
        dc.w    fetch-*         ; fetch character
        dc.w    send-*          ; send char (gives err.bp)

* open routine: check name and create channel if matched.          
          
open:
        move.w  (a0),d0
        subq.w  #4,d0
        blt.s   notf            ; must be at least 4 chars
        move.l  #$dfdfdfff,d0   ; ensure uppercase
        and.l   2(a0),d0
        cmpi.l  #'SRX1',d0      ; check against 'SRX1'
        beq.s   open2
        cmpi.l  #'STX1',d0
        bne.s   notf            ; exit if not
open2
        moveq   #$18,d1         ; just the bare minimum
        move.w  mm.alchp,a2
        jsr     (a2)            ; create channel definition block
        rts

notf    moveq   #err.nf,d0
        rts

* close: simply release channel def block          
          
close   move.w  mm.rechp,a2
        jmp     (a2)
          
* check for pending input
          
pend    moveq   #err.nc,d0      ; assume no pending input
        btst    #rxempty,status ; is RX fifo empty?
        bne.s   pendrts         ; yes, return NC
        moveq   #0,d0           ; else, return OK
pendrts rts

fetch   bsr     pend            ; fetch a byte if possible
        bne.s   fetchrts        ; but check if there is one first
        move.b  rxdata,d1       ; get byte out
fetchrts:
        rts                     ; return (D0 has been set already)

send    btst    #txempty,status ; is there room in the fifo
        beq.s   send_nc         ; no, return NC
        moveq   #8-1,d2         ; count 8 bits
        lea     txdata0,a1      ; NB: txdata1 is 1 higher
send_lp moveq   #1,d0
        and.b   d1,d0           ; keep only bit 0 of d1
        tst.b   (a1,d0.w)       ; read txdata0 or txdata1 as per bit 0
        ror.b   #1,d1           ; rotate data byte
        dbra    d2,send_lp      ; loop for 8 bits
        moveq   #0,d0           ; return OK
        rts
send_nc moveq   #err.nc,d0      ; TX fifo full, return NC
        rts

        end
